//---------------------------------------------------------------------------
//
//  FILE: cheetah.C
//  Author: Ken Barr
//  Date: Wed Nov  2 15:33:05 2005
//  Description:  A Pintool that feeds the Cheetah cache simulator
//  Copyright (C) 2005 Kenneth Barr (see COPYRIGHT file for details)
//
//---------------------------------------------------------------------------


#include <stdio.h>
#include "pin.H"
#include "libcheetah.h"
#include <getopt.h>

FILE * trace;

/* marshall up the libcheetah arguments */
int lib_argc = 0;
static char *lib_argv[16];

// my own args
int dcache=true;
LOCALVAR UINT64 maxinst=0;

UINT64 icount = 0;  //"count" is a reserved word!?
FILE * outfile;

// Print a memory read record
VOID RecordMemRead(VOID * ip, VOID * addr)
{
  cheetah_access((md_addr_t)addr);
  if(maxinst && ++icount > maxinst){
    PIN_RemoveFiniFunctions();
    PIN_Detach();
  }
}

// Print a memory write record
VOID RecordMemWrite(VOID * ip, VOID * addr)
{
    cheetah_access((md_addr_t)addr);
    if(maxinst && ++icount > maxinst){
      PIN_RemoveFiniFunctions();
      PIN_Detach();
    }
}


VOID icachetrace(INS ins, VOID *v)
{
  // Insert a call to printip before every instruction, and pass it the IP
    INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)RecordMemRead, IARG_INST_PTR, IARG_INST_PTR, IARG_END);

}

// Is called for every instruction and instruments reads and writes
VOID dcachetrace(INS ins, VOID *v)
{
    // instruments loads using a predicated call, i.e.
    // the call happens iff the load will be actually executed
    // (this does not matter for ia32 but arm and ipf have predicated instructions)
    if (INS_IsMemoryRead(ins))
    {
        INS_InsertPredicatedCall(
            ins, IPOINT_BEFORE, (AFUNPTR)RecordMemRead,
            IARG_INST_PTR,
            IARG_MEMORYREAD_EA,
            IARG_END);
    }

    // instruments stores using a predicated call, i.e.
    // the call happens iff the store will be actually executed
    if (INS_IsMemoryWrite(ins))
    {
        INS_InsertPredicatedCall(
            ins, IPOINT_BEFORE, (AFUNPTR)RecordMemWrite,
            IARG_INST_PTR,
            IARG_MEMORYWRITE_EA,
            IARG_END);
    }
}

VOID Fini(INT32 code, VOID *v)
{
  cheetah_stats(outfile, 0);
  //    fprintf(trace, "#eof\n");
  fclose(outfile);
}

VOID Detach(VOID *v)
{
  cheetah_stats(outfile, 0);
  //    fprintf(trace, "#eof\n");
  fclose(outfile);
}

int main(int argc, char *argv[])
{
  PIN_Init(argc, argv);

  //Pin doesn't care that these are cheetah args, but cheetah needs you to get rid of the pin args.
  argv=&argv[4];
  argc-=4;


  while(1){
    int c;
    int option_index = 0;
    char buf[512];
    static struct option long_options[] = {
      {"replacement", 1, 0, 'R'},
      {"configuration", 1, 0, 'C'},
      {"minsets", 1, 0, 'a'},
      {"maxsets", 1, 0, 'b'},
      {"linesize", 1,0,'l'},
      {"maxassoc", 1,0,'n'},
      {"maxsize", 1, 0, 'M'},
      {"cachesize", 1, 0, 'c'},
      {"interval", 1, 0, 'i'},
      {"maxinst",1,0,0},
      {"dcache",0,&dcache,1},
      {"icache",0,&dcache,0},
      {0,0,0,0}
    };

    c = getopt_long(argc, argv, "R:C:a:b:l:n:M:c:i:", long_options, &option_index);

    if (c == -1)
      break;

    switch(c){
       case 0:
        if(strcmp("maxinst", long_options[option_index].name)==0){
          maxinst = atoll(optarg);
        }
        else if(strcmp("dcache", long_options[option_index].name)==0){
          dcache = true;
        }
        else if(strcmp("icache", long_options[option_index].name)==0){
          dcache = false;
        }
        else{
          printf("Unrecognized option %s\n", long_options[option_index].name);
          exit(1);
        }
        break;
    case 'R':
    case 'C':
    case 'a':
    case 'b':
    case 'l':
    case 'n':
    case 'M':
    case 'i':
      sprintf(buf, "-%c%s", c, optarg);
      printf("copying %s\n", buf);
      lib_argv[lib_argc++] = strdup(buf);
      break;
    case '?':
      assert(0);
      break;

    default:
      printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }

  if (optind < argc) {
    printf ("non-option ARGV-elements: ");
    while (optind < argc)
      printf ("%s ", argv[optind++]);
    printf ("\n");
  }

/*sprintf(buf, "-i%d", cache_interval);
lib_argv[lib_argc++] = strdup(buf);

  sprintf(buf, "-M%d", max_cache);
  lib_argv[lib_argc++] = strdup(buf);

  sprintf(buf, "-c%d", cache_size);
  lib_argv[lib_argc++] = strdup(buf);
*/


  cheetah_init(lib_argc,lib_argv);
  outfile = fopen("data.out", "w");
  cheetah_config(outfile);

  if(dcache){
    printf("Tracing d-cache\n");
    INS_AddInstrumentFunction(dcachetrace, 0);
  }
  else{
    printf("Tracing i-cache\n");
     INS_AddInstrumentFunction(icachetrace, 0);
  }
  PIN_AddDetachFunction(Detach, 0);
  PIN_AddFiniFunction(Fini, 0);


  // Never returns
  PIN_StartProgram();

  return 0;
}
